
#define QT_VERSION
#define CUDA_KERNEL_INCLUDE
#include "../cellmask.h"

extern "C" __global__ void prepareCellPos(  float3 * out, int a, int b, int g,
                                            float aX, float aY, float aZ,
                                            float bX, float bY, float bZ,
                                            float gX, float gY, float gZ,
                                            float beginX, float beginY, float beginZ){
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  float3 c = make_float3(beginX + idx * aX, beginY + idx * aY, beginZ + idx * aZ);
  if(idx < a){
    idx = idx * b * g;
    __syncthreads();
    for(int i = 0 ; i < b ; i++){
      for(int j = 0 ; j < g ; j++){
        out[idx].x = c.x;
        out[idx].y = c.y;
        out[idx].z = c.z;
        __syncthreads();
        idx++;
        c.x += gX;
        c.y += gY;
        c.z += gZ;
        __syncthreads();
      }
      c.x += bX - gX * g;
      c.y += bY - gY * g;
      c.z += bZ - gZ * g;
      __syncthreads();
    }
    __syncthreads();
  }
  __syncthreads();
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void prepareCellMate(
    CellMate * out, int a, int BxG,
    int mate0, int mate1, int mate2,  int mate3,
    int mate4, int mate5, int mate6,  int mate7,
    int mate8, int mate9, int mate10, int mate11, int offset){
  int cidx, idx = blockIdx.x * blockDim.x + threadIdx.x;
  if(idx < a){
    idx = idx * BxG;
    cidx = idx + offset;
    __syncthreads();
    for(int i = 0 ; i < BxG ; i++){
      out[idx].m1[0]  = mate0  == 0 ? 0 : cidx + mate0;
      out[idx].m1[1]  = mate1  == 0 ? 0 : cidx + mate1;
      out[idx].m1[2]  = mate2  == 0 ? 0 : cidx + mate2;
      out[idx].m1[3]  = mate3  == 0 ? 0 : cidx + mate3;
      out[idx].m1[4]  = mate4  == 0 ? 0 : cidx + mate4;
      out[idx].m1[5]  = mate5  == 0 ? 0 : cidx + mate5;
      out[idx].m1[6]  = mate6  == 0 ? 0 : cidx + mate6;
      out[idx].m1[7]  = mate7  == 0 ? 0 : cidx + mate7;
      out[idx].m1[8]  = mate8  == 0 ? 0 : cidx + mate8;
      out[idx].m1[9]  = mate9  == 0 ? 0 : cidx + mate9;
      out[idx].m1[10] = mate10 == 0 ? 0 : cidx + mate10;
      out[idx].m1[11] = mate11 == 0 ? 0 : cidx + mate11;
      __syncthreads();
      idx++; cidx++;
      __syncthreads();
    }
    __syncthreads();
  }
  __syncthreads();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void prepareCellMask(float3 * pos, unsigned long * outMask, int a, int BxG, unsigned long maskBase,
                                           float wx0, float wx1, float wy0, float wy1, float wz0, float wz1)
{
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  float3 c1;

  if(idx < a){
    idx = idx * BxG;
    __syncthreads();
    for(int i = 0 ; i < BxG ; i++){
      c1.x = pos[idx].x;
      c1.y = pos[idx].y;
      c1.z = pos[idx].z;
      __syncthreads();
      if( c1.x < wx0 || c1.x > wx1 ||
          c1.y < wy0 || c1.y > wy1 ||
          c1.z < wz0 || c1.z > wz1){
        outMask[idx] = CELL_MASK_NULL;
      } else {
        outMask[idx] = maskBase;
      }
      __syncthreads();
      idx++;
      __syncthreads();
    }
    __syncthreads();
  }
  __syncthreads();
}

/*

extern __shared__ TriangF triang[];

extern "C" __global__ void PrepareCellAttr( float3 * pos, CellAttribute * attr, TriangF * triangs, int cntTriangs,
                                            int a, int BxG, int grainIdx, int type, float maxDist, unsigned long maskBase, unsigned long maskBaseBord){
  int idx = blockIdx.x * blockDim.x + threadIdx.x, cntOut, cntDist;
  float dist;
  float3 c1;

  if(idx < cntTriangs){
    triang[idx].nx  = triangs[idx].nx;
    triang[idx].ny  = triangs[idx].ny;
    triang[idx].nz  = triangs[idx].nz;
    triang[idx].d   = triangs[idx].d;
  }
  __syncthreads();

  if(idx < a){
    idx = idx * BxG;
    __syncthreads();
    for(int i = 0 ; i < BxG ; i++){
      cntOut  = 0;
      cntDist = 0;
      c1.x = pos[idx].x;
      c1.y = pos[idx].y;
      c1.z = pos[idx].z;
      __syncthreads();
      for(int j = 0 ; j < cntTriangs; j++){
        dist = triang[j].nx*c1.x + triang[j].ny*c1.y + triang[j].nz*c1.z + triang[j].d;
        if(dist < 0.0f)cntOut++;
        if(dist < maxDist)cntDist++;
      }
      __syncthreads();
      if(cntOut == 0){
        attr[idx].grainIdx = grainIdx;
        attr[idx].type     = type;
        if(cntDist > 0){
          attr[idx].mask   = maskBaseBord;
        }else{
          attr[idx].mask   = maskBase;
        }
      }else{
        attr[idx].grainIdx = 0;
        attr[idx].type     = 0;
        attr[idx].mask     = CELL_MASK_NULL;
      }
      __syncthreads();
      idx++;
      __syncthreads();
    }
    __syncthreads();
  }
  __syncthreads();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern "C" __global__ void PrepareBorderCells(int * outID, float3 * pos, CellAttribute * attr, TriangF * triangs, int cntTriangs, int cellCount, int steps, float maxDist){
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int outIdx, cntDist;
  float3 c1;

  if(idx < cntTriangs){
    triang[idx].nx  = triangs[idx].nx;
    triang[idx].ny  = triangs[idx].ny;
    triang[idx].nz  = triangs[idx].nz;
    triang[idx].d   = triangs[idx].d;
  }

  idx = idx * steps;
  outIdx = idx;

  for(int i = 0 ; i < steps ; i++){
    if(idx < cellCount && __IS_CELL_MASK_PART_DB_BORDER(attr[idx].mask)){
      cntDist = 0;
      c1.x = pos[idx].x;
      c1.y = pos[idx].y;
      c1.z = pos[idx].z;
      __syncthreads();
      for(int j = 0 ; j < cntTriangs; j++){
        if((triang[j].nx*c1.x + triang[j].ny*c1.y + triang[j].nz*c1.z + triang[j].d) < maxDist)cntDist++;
      }
      __syncthreads();
      if(cntDist > 0){
        __SET_CELL_MASK_SCENE_BORDER(attr[idx].mask);
        outID[outIdx] = idx;
        outIdx++;
      }
      __syncthreads();
    }
    idx++;
    __syncthreads();
  }
  __syncthreads();
}
*/
