#ifndef ___PROCES_F_CELL_DATA_H___
#define ___PROCES_F_CELL_DATA_H___

#ifndef CUDA_KERNEL_INCLUDE
  #define CUDA_KERNEL_INCLUDE
  #define QT_VERSION
#endif

#include "../cellmask.h"

extern __shared__ CalcDataF cellData[];

__device__ void StoreCalcDataF(int idx, CalcDataF * data){
  cellData[idx].isVac   = data[idx].isVac;  /*cellData[idx].tempr   = data[idx].tempr;    cellData[idx].id    = data[idx].id;*/
  cellData[idx].subType = data[idx].subType;  cellData[idx].intType = data[idx].intType;  cellData[idx].crStr = data[idx].crStr;
  cellData[idx].lA      = data[idx].lA;       cellData[idx].lB      = data[idx].lB;       cellData[idx].d     = data[idx].d;
  cellData[idx].aX      = data[idx].aX;       cellData[idx].aY      = data[idx].aY;       cellData[idx].aZ    = data[idx].aZ;
  cellData[idx].bX      = data[idx].bX;       cellData[idx].bY      = data[idx].bY;       cellData[idx].bZ    = data[idx].bZ;
  cellData[idx].gX      = data[idx].gX;       cellData[idx].gY      = data[idx].gY;       cellData[idx].gZ    = data[idx].gZ;
}

__device__ int GetCalcDataIDX(bool isVac, /*short tempr, int grainIdx,*/ int subType, int intType, CrystalMatrixType crStr, int dataCount){
  for(int i = 0 ; i < dataCount ; i++){
    if(cellData[i].isVac    == isVac    &&
     //cellData[i].tempr    == tempr    &&
     //cellData[i].id       == grainIdx &&
       cellData[i].subType  == subType  &&
       cellData[i].intType  == intType  &&
       cellData[i].crStr    == crStr)return i;
  }
  return -1;
}

__device__ int IsNearInterstitial(CellAttribute * attr, int mateId/*, short curTempr, int grainIdx*/, CrystalMatrixType crStr, int dataCount){
  int mateType;
  unsigned long mateMask = attr[mateId].mask;
  __syncthreads();
  if(__IS_CELL_MASK_BUG_INT_ANY(mateMask)){
    mateType = attr[mateId].type;
  __syncthreads();
    return GetCalcDataIDX(false/*, curTempr, grainIdx*/, -1, mateType, crStr, dataCount);
  }
  else return -1;
}

__device__ __inline__ void SetLaticePositionsBCC2(float3 * out, float3 * pos, int dir, int dataID){
  SetLaticePositionsBCC(dir, out, pos,
                        cellData[dataID].aX, cellData[dataID].aY, cellData[dataID].aZ,
                        cellData[dataID].bX, cellData[dataID].bY, cellData[dataID].bZ,
                        cellData[dataID].gX, cellData[dataID].gY, cellData[dataID].gZ);
  __syncthreads();
}

__device__ __inline__ void SetLaticePositionsFCC2(float3 * out, float3 * pos, int dir, int dataID){
  SetLaticePositionsFCC(dir, out, pos,
                        cellData[dataID].aX, cellData[dataID].aY, cellData[dataID].aZ,
                        cellData[dataID].bX, cellData[dataID].bY, cellData[dataID].bZ,
                        cellData[dataID].gX, cellData[dataID].gY, cellData[dataID].gZ);
  __syncthreads();
}

#endif
