#ifndef ___CUDA_CELL_MASK_H___
#define ___CUDA_CELL_MASK_H___

#ifdef CUDA_KERNEL_INCLUDE
  #define __USE__DEVICE__                 __device__
  #define __USE__INLINE__                 __inline__
#else
  #define __USE__DEVICE__
  #define __USE__INLINE__                 inline
  #include <QtCore>
  #include <QtXml>
  #include <float.h>
#endif

#include <vector_functions.h>


enum CrystalMatrixType
{
  BCC, FCC
};

enum BugType
{
  VACANCY,
//  INTERSTITIAL,
  SUBSTITUTION
};

enum BCCDirections {
  PA_PB_PG  = 0,
  MA_MB_MG  = 1,
  PA_PB_MG  = 2,
  MA_MB_PG  = 3,
  PA_MB_PG  = 4,
  MA_PB_MG  = 5,
  PA_MB_MG  = 6,
  MA_PB_PG  = 7
};

enum FCCDirections {
  PA_PB  = 0,
  MA_MB  = 1,
  PA_MB  = 2,
  MA_PB  = 3,
  PA_PG  = 4,
  MA_MG  = 5,
  PA_MG  = 6,
  MA_PG  = 7,
  PB_PG  = 8,
  MB_MG  = 9,
  PB_MG  = 10,
  MB_PG  = 11
};

#define MAX_MATES       12
#define MAX_MATESx2     24
//#define MAX_MATESx3     36

#define CNT_DIR_BCC     8
#define CNT_DIR_FCC     12

struct CellMate { int m1[MAX_MATES], m2[MAX_MATES]; };

struct CellAttribute {
  int           type/*, grainIdx*/;
  unsigned long mask;
};

struct CalcDataF {
  bool              isVac;
//  short             tempr;
  int               /*id, */subType, intType;
  CrystalMatrixType crStr;
  float             lA, lB, d;
  float             aX, aY, aZ;
  float             bX, bY, bZ;
  float             gX, gY, gZ;
};

/** ********************************************************************************************************** */
/*  ********************************************************************************************************** */
/** ********************************************************************************************************** */

#define CELL_LATTICE_CNT                        8

#define CELL_MASK_NULL                          0X0000000000000000

#define CELL_MASK_LATTICE_1                     0X0100000000000000
#define CELL_MASK_LATTICE_2                     0X0200000000000000
#define CELL_MASK_LATTICE_3                     0X0400000000000000
#define CELL_MASK_LATTICE_4                     0X0800000000000000
#define CELL_MASK_LATTICE_5                     0X1000000000000000
#define CELL_MASK_LATTICE_6                     0X2000000000000000
#define CELL_MASK_LATTICE_7                     0X4000000000000000
#define CELL_MASK_LATTICE_8                     0X8000000000000000

#define CELL_MASK_LATTICE_1_NEG                 0XFEFFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_2_NEG                 0XFDFFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_3_NEG                 0XFBFFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_4_NEG                 0XF7FFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_5_NEG                 0XEFFFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_6_NEG                 0XDFFFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_7_NEG                 0XBFFFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_8_NEG                 0X7FFFFFFFFFFFFFFF

#define CELL_MASK_LATTICE_RESET                 0X00FFFFFFFFFFFFFF
#define CELL_MASK_LATTICE_RESET_NEG             0XFF00000000000000

#define __SET_CELL_MASK_LATTICE_1(x)            ((x) |= CELL_MASK_LATTICE_1)
#define __SET_CELL_MASK_LATTICE_2(x)            ((x) |= CELL_MASK_LATTICE_2)
#define __SET_CELL_MASK_LATTICE_3(x)            ((x) |= CELL_MASK_LATTICE_3)
#define __SET_CELL_MASK_LATTICE_4(x)            ((x) |= CELL_MASK_LATTICE_4)
#define __SET_CELL_MASK_LATTICE_5(x)            ((x) |= CELL_MASK_LATTICE_5)
#define __SET_CELL_MASK_LATTICE_6(x)            ((x) |= CELL_MASK_LATTICE_6)
#define __SET_CELL_MASK_LATTICE_7(x)            ((x) |= CELL_MASK_LATTICE_7)
#define __SET_CELL_MASK_LATTICE_8(x)            ((x) |= CELL_MASK_LATTICE_8)

#define __UNSET_CELL_MASK_LATTICE_1(x)          ((x) &= CELL_MASK_LATTICE_1_NEG)
#define __UNSET_CELL_MASK_LATTICE_2(x)          ((x) &= CELL_MASK_LATTICE_2_NEG)
#define __UNSET_CELL_MASK_LATTICE_3(x)          ((x) &= CELL_MASK_LATTICE_3_NEG)
#define __UNSET_CELL_MASK_LATTICE_4(x)          ((x) &= CELL_MASK_LATTICE_4_NEG)
#define __UNSET_CELL_MASK_LATTICE_5(x)          ((x) &= CELL_MASK_LATTICE_5_NEG)
#define __UNSET_CELL_MASK_LATTICE_6(x)          ((x) &= CELL_MASK_LATTICE_6_NEG)
#define __UNSET_CELL_MASK_LATTICE_7(x)          ((x) &= CELL_MASK_LATTICE_7_NEG)
#define __UNSET_CELL_MASK_LATTICE_8(x)          ((x) &= CELL_MASK_LATTICE_8_NEG)

#define __RESET_CELL_LATTICE(x)                 ((x) &= CELL_MASK_LATTICE_RESET)

#define __IS_CELL_MASK_LATTICE_1(x)             ((x) & CELL_MASK_LATTICE_1)
#define __IS_CELL_MASK_LATTICE_2(x)             ((x) & CELL_MASK_LATTICE_2)
#define __IS_CELL_MASK_LATTICE_3(x)             ((x) & CELL_MASK_LATTICE_3)
#define __IS_CELL_MASK_LATTICE_4(x)             ((x) & CELL_MASK_LATTICE_4)
#define __IS_CELL_MASK_LATTICE_5(x)             ((x) & CELL_MASK_LATTICE_5)
#define __IS_CELL_MASK_LATTICE_6(x)             ((x) & CELL_MASK_LATTICE_6)
#define __IS_CELL_MASK_LATTICE_7(x)             ((x) & CELL_MASK_LATTICE_7)
#define __IS_CELL_MASK_LATTICE_8(x)             ((x) & CELL_MASK_LATTICE_8)

__USE__DEVICE__ __USE__INLINE__ void CellMaskSetLattice(unsigned long* mask, int latice){
  __RESET_CELL_LATTICE(*mask);
  switch(latice){
    case(1):__SET_CELL_MASK_LATTICE_1(*mask); return;
    case(2):__SET_CELL_MASK_LATTICE_2(*mask); return;
    case(3):__SET_CELL_MASK_LATTICE_3(*mask); return;
    case(4):__SET_CELL_MASK_LATTICE_4(*mask); return;
    case(5):__SET_CELL_MASK_LATTICE_5(*mask); return;
    case(6):__SET_CELL_MASK_LATTICE_6(*mask); return;
    case(7):__SET_CELL_MASK_LATTICE_7(*mask); return;
    case(8):__SET_CELL_MASK_LATTICE_8(*mask); return;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetLattice(unsigned long mask){
  switch(CELL_MASK_LATTICE_RESET_NEG & mask){
    case(CELL_MASK_LATTICE_1): return 1;
    case(CELL_MASK_LATTICE_2): return 2;
    case(CELL_MASK_LATTICE_3): return 3;
    case(CELL_MASK_LATTICE_4): return 4;
    case(CELL_MASK_LATTICE_5): return 5;
    case(CELL_MASK_LATTICE_6): return 6;
    case(CELL_MASK_LATTICE_7): return 7;
    case(CELL_MASK_LATTICE_8): return 8;
  default: return 0;
  }
}

/** ********************************************************************************************************** */
/*  ********************************************************************************************************** */
/** ********************************************************************************************************** */

#define CELL_MASK_SYSTEM_BCC                    0X0001000000000000
#define CELL_MASK_SYSTEM_FCC                    0X0002000000000000
#define CELL_MASK_SYSTEM_LIQUID                 0X0004000000000000
#define CELL_MASK_SYSTEM_UNDEF1                 0X0008000000000000
#define CELL_MASK_SYSTEM_UNDEF2                 0X0010000000000000
#define CELL_MASK_SYSTEM_UNDEF3                 0X0020000000000000
#define CELL_MASK_SYSTEM_UNDEF4                 0X0040000000000000
#define CELL_MASK_SYSTEM_UNDEF5                 0X0080000000000000

#define CELL_MASK_SYSTEM_BCC_NEG                0XFFFEFFFFFFFFFFFF
#define CELL_MASK_SYSTEM_FCC_NEG                0XFFFDFFFFFFFFFFFF
#define CELL_MASK_SYSTEM_LIQUID_NEG             0XFFFBFFFFFFFFFFFF
#define CELL_MASK_SYSTEM_UNDEF1_NEG             0XFFF7FFFFFFFFFFFF
#define CELL_MASK_SYSTEM_UNDEF2_NEG             0XFFEFFFFFFFFFFFFF
#define CELL_MASK_SYSTEM_UNDEF3_NEG             0XFFDFFFFFFFFFFFFF
#define CELL_MASK_SYSTEM_UNDEF4_NEG             0XFFBFFFFFFFFFFFFF
#define CELL_MASK_SYSTEM_UNDEF5_NEG             0XFF7FFFFFFFFFFFFF

#define CELL_MASK_SYSTEM_RESET                  0XFF00FFFFFFFFFFFF
#define CELL_MASK_SYSTEM_RESET_NEG              0X00FF000000000000

#define __SET_CELL_MASK_SYSTEM_BCC(x)           ((x) |= CELL_MASK_SYSTEM_BCC)
#define __SET_CELL_MASK_SYSTEM_FCC(x)           ((x) |= CELL_MASK_SYSTEM_FCC)
//#define __SET_CELL_MASK_SYSTEM_LIQUID(x)        ((x) |= CELL_MASK_SYSTEM_LIQUID)

#define __UNSET_CELL_MASK_SYSTEM_BCC(x)         ((x) &= CELL_MASK_SYSTEM_BCC_NEG)
#define __UNSET_CELL_MASK_SYSTEM_FCC(x)         ((x) &= CELL_MASK_SYSTEM_FCC_NEG)
//#define __UNSET_CELL_MASK_SYSTEM_LIQUID(x)      ((x) &= CELL_MASK_SYSTEM_LIQUID_NEG)

#define __RESET_CELL_SYSTEM(x)                  ((x) &= CELL_MASK_SYSTEM_RESET)

#define __IS_CELL_MASK_SYSTEM_BCC(x)            ((x) & CELL_MASK_SYSTEM_BCC)
#define __IS_CELL_MASK_SYSTEM_FCC(x)            ((x) & CELL_MASK_SYSTEM_FCC)
//#define __IS_CELL_MASK_SYSTEM_LIQUID(x)         ((x) & CELL_MASK_SYSTEM_LIQUID)

__USE__DEVICE__ __USE__INLINE__ void CellMaskSetCrystalStruct(unsigned long* mask, CrystalMatrixType system){
  __RESET_CELL_SYSTEM(*mask);
  switch(system){
    case(BCC):    __SET_CELL_MASK_SYSTEM_BCC(*mask); return;
    case(FCC):    __SET_CELL_MASK_SYSTEM_FCC(*mask); return;
  }
}

__USE__DEVICE__ __USE__INLINE__ CrystalMatrixType CellMaskGetCrystalStruct(unsigned long mask){
  switch(CELL_MASK_SYSTEM_RESET_NEG & mask){
    case(CELL_MASK_SYSTEM_BCC): return BCC;
    case(CELL_MASK_SYSTEM_FCC): return FCC;
  default: return BCC;
  }
}

/** ********************************************************************************************************** */
/*  ********************************************************************************************************** */
/** ********************************************************************************************************** */

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_0_BIT_1                  0X0000000000000001
#define CELL_MASK_MATE_0_BIT_2                  0X0000000000000002
#define CELL_MASK_MATE_0_BIT_1_NEG              0XFFFFFFFFFFFFFFFE
#define CELL_MASK_MATE_0_BIT_2_NEG              0XFFFFFFFFFFFFFFFD

#define CELL_MASK_MATE_0_RESET                  0XFFFFFFFFFFFFFFFC
#define CELL_MASK_MATE_0_RESET_NEG              0X0000000000000003

#define __UNSET_CELL_MASK_MATE_0(x)             (x) &= CELL_MASK_MATE_0_RESET
#define __SET_CELL_MASK_MATE_0_SET_1(x)         ((x) |= CELL_MASK_MATE_0_BIT_1)    ; ((x) &= CELL_MASK_MATE_0_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_0_SET_2(x)         ((x) &= CELL_MASK_MATE_0_BIT_1_NEG); ((x) |= CELL_MASK_MATE_0_BIT_2)
#define __SET_CELL_MASK_MATE_0_SET_3(x)         ((x) |= CELL_MASK_MATE_0_BIT_1)    ; ((x) |= CELL_MASK_MATE_0_BIT_2)

#define __IS_CELL_MASK_MATE_0_BIT_1(x)          ((x) & CELL_MASK_MATE_0_BIT_1)
#define __IS_CELL_MASK_MATE_0_BIT_2(x)          ((x) & CELL_MASK_MATE_0_BIT_2)
#define __IS_CELL_MASK_MATE_0(x)                (__IS_CELL_MASK_MATE_0_BIT_1((x)) || __IS_CELL_MASK_MATE_0_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_1_BIT_1                  0X0000000000000004
#define CELL_MASK_MATE_1_BIT_2                  0X0000000000000008
#define CELL_MASK_MATE_1_BIT_1_NEG              0XFFFFFFFFFFFFFFFB
#define CELL_MASK_MATE_1_BIT_2_NEG              0XFFFFFFFFFFFFFFF7

#define CELL_MASK_MATE_1_RESET                  0XFFFFFFFFFFFFFFF3
#define CELL_MASK_MATE_1_RESET_NEG              0X000000000000000C

#define __UNSET_CELL_MASK_MATE_1(x)             ((x) &= CELL_MASK_MATE_1_RESET)
#define __SET_CELL_MASK_MATE_1_SET_1(x)         ((x) |= CELL_MASK_MATE_1_BIT_1)    ; ((x) &= CELL_MASK_MATE_1_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_1_SET_2(x)         ((x) &= CELL_MASK_MATE_1_BIT_1_NEG); ((x) |= CELL_MASK_MATE_1_BIT_2)
#define __SET_CELL_MASK_MATE_1_SET_3(x)         ((x) |= CELL_MASK_MATE_1_BIT_1)    ; ((x) |= CELL_MASK_MATE_1_BIT_2)

#define __IS_CELL_MASK_MATE_1_BIT_1(x)          ((x) & CELL_MASK_MATE_1_BIT_1)
#define __IS_CELL_MASK_MATE_1_BIT_2(x)          ((x) & CELL_MASK_MATE_1_BIT_2)
#define __IS_CELL_MASK_MATE_1(x)                (__IS_CELL_MASK_MATE_1_BIT_1((x)) || __IS_CELL_MASK_MATE_1_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_2_BIT_1                  0X0000000000000010
#define CELL_MASK_MATE_2_BIT_2                  0X0000000000000020
#define CELL_MASK_MATE_2_BIT_1_NEG              0XFFFFFFFFFFFFFFEF
#define CELL_MASK_MATE_2_BIT_2_NEG              0XFFFFFFFFFFFFFFDF

#define CELL_MASK_MATE_2_RESET                  0XFFFFFFFFFFFFFFCF
#define CELL_MASK_MATE_2_RESET_NEG              0X0000000000000030

#define __UNSET_CELL_MASK_MATE_2(x)             ((x) &= CELL_MASK_MATE_2_RESET)
#define __SET_CELL_MASK_MATE_2_SET_1(x)         ((x) |= CELL_MASK_MATE_2_BIT_1)    ; ((x) &= CELL_MASK_MATE_2_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_2_SET_2(x)         ((x) &= CELL_MASK_MATE_2_BIT_1_NEG); ((x) |= CELL_MASK_MATE_2_BIT_2)
#define __SET_CELL_MASK_MATE_2_SET_3(x)         ((x) |= CELL_MASK_MATE_2_BIT_1)    ; ((x) |= CELL_MASK_MATE_2_BIT_2)

#define __IS_CELL_MASK_MATE_2_BIT_1(x)          ((x) & CELL_MASK_MATE_2_BIT_1)
#define __IS_CELL_MASK_MATE_2_BIT_2(x)          ((x) & CELL_MASK_MATE_2_BIT_2)
#define __IS_CELL_MASK_MATE_2(x)                (__IS_CELL_MASK_MATE_2_BIT_1((x)) || __IS_CELL_MASK_MATE_2_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_3_BIT_1                  0X0000000000000040
#define CELL_MASK_MATE_3_BIT_2                  0X0000000000000080
#define CELL_MASK_MATE_3_BIT_1_NEG              0XFFFFFFFFFFFFFFBF
#define CELL_MASK_MATE_3_BIT_2_NEG              0XFFFFFFFFFFFFFF7F

#define CELL_MASK_MATE_3_RESET                  0XFFFFFFFFFFFFFF3F
#define CELL_MASK_MATE_3_RESET_NEG              0X00000000000000C0

#define __UNSET_CELL_MASK_MATE_3(x)             ((x) &= CELL_MASK_MATE_3_RESET)
#define __SET_CELL_MASK_MATE_3_SET_1(x)         ((x) |= CELL_MASK_MATE_3_BIT_1)    ; ((x) &= CELL_MASK_MATE_3_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_3_SET_2(x)         ((x) &= CELL_MASK_MATE_3_BIT_1_NEG); ((x) |= CELL_MASK_MATE_3_BIT_2)
#define __SET_CELL_MASK_MATE_3_SET_3(x)         ((x) |= CELL_MASK_MATE_3_BIT_1)    ; ((x) |= CELL_MASK_MATE_3_BIT_2)

#define __IS_CELL_MASK_MATE_3_BIT_1(x)          ((x) & CELL_MASK_MATE_3_BIT_1)
#define __IS_CELL_MASK_MATE_3_BIT_2(x)          ((x) & CELL_MASK_MATE_3_BIT_2)
#define __IS_CELL_MASK_MATE_3(x)                (__IS_CELL_MASK_MATE_3_BIT_1((x)) || __IS_CELL_MASK_MATE_3_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_4_BIT_1                  0X0000000000000100
#define CELL_MASK_MATE_4_BIT_2                  0X0000000000000200
#define CELL_MASK_MATE_4_BIT_1_NEG              0XFFFFFFFFFFFFFEFF
#define CELL_MASK_MATE_4_BIT_2_NEG              0XFFFFFFFFFFFFFDFF

#define CELL_MASK_MATE_4_RESET                  0XFFFFFFFFFFFFFCFF
#define CELL_MASK_MATE_4_RESET_NEG              0X0000000000000300

#define __UNSET_CELL_MASK_MATE_4(x)             ((x) &= CELL_MASK_MATE_4_RESET)
#define __SET_CELL_MASK_MATE_4_SET_1(x)         ((x) |= CELL_MASK_MATE_4_BIT_1)    ; ((x) &= CELL_MASK_MATE_4_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_4_SET_2(x)         ((x) &= CELL_MASK_MATE_4_BIT_1_NEG); ((x) |= CELL_MASK_MATE_4_BIT_2)
#define __SET_CELL_MASK_MATE_4_SET_3(x)         ((x) |= CELL_MASK_MATE_4_BIT_1)    ; ((x) |= CELL_MASK_MATE_4_BIT_2)

#define __IS_CELL_MASK_MATE_4_BIT_1(x)          ((x) & CELL_MASK_MATE_4_BIT_1)
#define __IS_CELL_MASK_MATE_4_BIT_2(x)          ((x) & CELL_MASK_MATE_4_BIT_2)
#define __IS_CELL_MASK_MATE_4(x)                (__IS_CELL_MASK_MATE_4_BIT_1((x)) || __IS_CELL_MASK_MATE_4_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_5_BIT_1                  0X0000000000000400
#define CELL_MASK_MATE_5_BIT_2                  0X0000000000000800
#define CELL_MASK_MATE_5_BIT_1_NEG              0XFFFFFFFFFFFFFBFF
#define CELL_MASK_MATE_5_BIT_2_NEG              0XFFFFFFFFFFFFF7FF

#define CELL_MASK_MATE_5_RESET                  0XFFFFFFFFFFFFF3FF
#define CELL_MASK_MATE_5_RESET_NEG              0X0000000000000C00

#define __UNSET_CELL_MASK_MATE_5(x)             ((x) &= CELL_MASK_MATE_5_RESET)
#define __SET_CELL_MASK_MATE_5_SET_1(x)         ((x) |= CELL_MASK_MATE_5_BIT_1)    ; ((x) &= CELL_MASK_MATE_5_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_5_SET_2(x)         ((x) &= CELL_MASK_MATE_5_BIT_1_NEG); ((x) |= CELL_MASK_MATE_5_BIT_2)
#define __SET_CELL_MASK_MATE_5_SET_3(x)         ((x) |= CELL_MASK_MATE_5_BIT_1)    ; ((x) |= CELL_MASK_MATE_5_BIT_2)

#define __IS_CELL_MASK_MATE_5_BIT_1(x)          ((x) & CELL_MASK_MATE_5_BIT_1)
#define __IS_CELL_MASK_MATE_5_BIT_2(x)          ((x) & CELL_MASK_MATE_5_BIT_2)
#define __IS_CELL_MASK_MATE_5(x)                (__IS_CELL_MASK_MATE_5_BIT_1((x)) || __IS_CELL_MASK_MATE_5_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_6_BIT_1                  0X0000000000001000
#define CELL_MASK_MATE_6_BIT_2                  0X0000000000002000
#define CELL_MASK_MATE_6_BIT_1_NEG              0XFFFFFFFFFFFFEFFF
#define CELL_MASK_MATE_6_BIT_2_NEG              0XFFFFFFFFFFFFDFFF

#define CELL_MASK_MATE_6_RESET                  0XFFFFFFFFFFFFCFFF
#define CELL_MASK_MATE_6_RESET_NEG              0X0000000000003000

#define __UNSET_CELL_MASK_MATE_6(x)             ((x) &= CELL_MASK_MATE_6_RESET)
#define __SET_CELL_MASK_MATE_6_SET_1(x)         ((x) |= CELL_MASK_MATE_6_BIT_1)    ; ((x) &= CELL_MASK_MATE_6_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_6_SET_2(x)         ((x) &= CELL_MASK_MATE_6_BIT_1_NEG); ((x) |= CELL_MASK_MATE_6_BIT_2)
#define __SET_CELL_MASK_MATE_6_SET_3(x)         ((x) |= CELL_MASK_MATE_6_BIT_1)    ; ((x) |= CELL_MASK_MATE_6_BIT_2)

#define __IS_CELL_MASK_MATE_6_BIT_1(x)          ((x) & CELL_MASK_MATE_6_BIT_1)
#define __IS_CELL_MASK_MATE_6_BIT_2(x)          ((x) & CELL_MASK_MATE_6_BIT_2)
#define __IS_CELL_MASK_MATE_6(x)                (__IS_CELL_MASK_MATE_6_BIT_1((x)) || __IS_CELL_MASK_MATE_6_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_7_BIT_1                  0X0000000000004000
#define CELL_MASK_MATE_7_BIT_2                  0X0000000000008000
#define CELL_MASK_MATE_7_BIT_1_NEG              0XFFFFFFFFFFFFBFFF
#define CELL_MASK_MATE_7_BIT_2_NEG              0XFFFFFFFFFFFF7FFF

#define CELL_MASK_MATE_7_RESET                  0XFFFFFFFFFFFF3FFF
#define CELL_MASK_MATE_7_RESET_NEG              0X000000000000C000

#define __UNSET_CELL_MASK_MATE_7(x)             ((x) &= CELL_MASK_MATE_7_RESET)
#define __SET_CELL_MASK_MATE_7_SET_1(x)         ((x) |= CELL_MASK_MATE_7_BIT_1)    ; ((x) &= CELL_MASK_MATE_7_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_7_SET_2(x)         ((x) &= CELL_MASK_MATE_7_BIT_1_NEG); ((x) |= CELL_MASK_MATE_7_BIT_2)
#define __SET_CELL_MASK_MATE_7_SET_3(x)         ((x) |= CELL_MASK_MATE_7_BIT_1)    ; ((x) |= CELL_MASK_MATE_7_BIT_2)

#define __IS_CELL_MASK_MATE_7_BIT_1(x)          ((x) & CELL_MASK_MATE_7_BIT_1)
#define __IS_CELL_MASK_MATE_7_BIT_2(x)          ((x) & CELL_MASK_MATE_7_BIT_2)
#define __IS_CELL_MASK_MATE_7(x)                (__IS_CELL_MASK_MATE_7_BIT_1((x)) || __IS_CELL_MASK_MATE_7_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_8_BIT_1                  0X0000000000010000
#define CELL_MASK_MATE_8_BIT_2                  0X0000000000020000
#define CELL_MASK_MATE_8_BIT_1_NEG              0XFFFFFFFFFFFEFFFF
#define CELL_MASK_MATE_8_BIT_2_NEG              0XFFFFFFFFFFFDFFFF

#define CELL_MASK_MATE_8_RESET                  0XFFFFFFFFFFFCFFFF
#define CELL_MASK_MATE_8_RESET_NEG              0X0000000000030000

#define __UNSET_CELL_MASK_MATE_8(x)             ((x) &= CELL_MASK_MATE_8_RESET)
#define __SET_CELL_MASK_MATE_8_SET_1(x)         ((x) |= CELL_MASK_MATE_8_BIT_1)    ; ((x) &= CELL_MASK_MATE_8_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_8_SET_2(x)         ((x) &= CELL_MASK_MATE_8_BIT_1_NEG); ((x) |= CELL_MASK_MATE_8_BIT_2)
#define __SET_CELL_MASK_MATE_8_SET_3(x)         ((x) |= CELL_MASK_MATE_8_BIT_1)    ; ((x) |= CELL_MASK_MATE_8_BIT_2)

#define __IS_CELL_MASK_MATE_8_BIT_1(x)          ((x) & CELL_MASK_MATE_8_BIT_1)
#define __IS_CELL_MASK_MATE_8_BIT_2(x)          ((x) & CELL_MASK_MATE_8_BIT_2)
#define __IS_CELL_MASK_MATE_8(x)                (__IS_CELL_MASK_MATE_8_BIT_1((x)) || __IS_CELL_MASK_MATE_8_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_9_BIT_1                  0X0000000000040000
#define CELL_MASK_MATE_9_BIT_2                  0X0000000000080000
#define CELL_MASK_MATE_9_BIT_1_NEG              0XFFFFFFFFFFFBFFFF
#define CELL_MASK_MATE_9_BIT_2_NEG              0XFFFFFFFFFFF7FFFF

#define CELL_MASK_MATE_9_RESET                  0XFFFFFFFFFFF3FFFF
#define CELL_MASK_MATE_9_RESET_NEG              0X00000000000C0000

#define __UNSET_CELL_MASK_MATE_9(x)             ((x) &= CELL_MASK_MATE_9_RESET)
#define __SET_CELL_MASK_MATE_9_SET_1(x)         ((x) |= CELL_MASK_MATE_9_BIT_1)    ; ((x) &= CELL_MASK_MATE_9_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_9_SET_2(x)         ((x) &= CELL_MASK_MATE_9_BIT_1_NEG); ((x) |= CELL_MASK_MATE_9_BIT_2)
#define __SET_CELL_MASK_MATE_9_SET_3(x)         ((x) |= CELL_MASK_MATE_9_BIT_1)    ; ((x) |= CELL_MASK_MATE_9_BIT_2)

#define __IS_CELL_MASK_MATE_9_BIT_1(x)          ((x) & CELL_MASK_MATE_9_BIT_1)
#define __IS_CELL_MASK_MATE_9_BIT_2(x)          ((x) & CELL_MASK_MATE_9_BIT_2)
#define __IS_CELL_MASK_MATE_9(x)                (__IS_CELL_MASK_MATE_9_BIT_1((x)) || __IS_CELL_MASK_MATE_9_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_10_BIT_1                  0X0000000000100000
#define CELL_MASK_MATE_10_BIT_2                  0X0000000000200000
#define CELL_MASK_MATE_10_BIT_1_NEG              0XFFFFFFFFFFEFFFFF
#define CELL_MASK_MATE_10_BIT_2_NEG              0XFFFFFFFFFFDFFFFF

#define CELL_MASK_MATE_10_RESET                  0XFFFFFFFFFFCFFFFF
#define CELL_MASK_MATE_10_RESET_NEG              0X0000000000300000

#define __UNSET_CELL_MASK_MATE_10(x)             ((x) &= CELL_MASK_MATE_10_RESET)
#define __SET_CELL_MASK_MATE_10_SET_1(x)         ((x) |= CELL_MASK_MATE_10_BIT_1)    ; ((x) &= CELL_MASK_MATE_10_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_10_SET_2(x)         ((x) &= CELL_MASK_MATE_10_BIT_1_NEG); ((x) |= CELL_MASK_MATE_10_BIT_2)
#define __SET_CELL_MASK_MATE_10_SET_3(x)         ((x) |= CELL_MASK_MATE_10_BIT_1)    ; ((x) |= CELL_MASK_MATE_10_BIT_2)

#define __IS_CELL_MASK_MATE_10_BIT_1(x)          ((x) & CELL_MASK_MATE_10_BIT_1)
#define __IS_CELL_MASK_MATE_10_BIT_2(x)          ((x) & CELL_MASK_MATE_10_BIT_2)
#define __IS_CELL_MASK_MATE_10(x)                (__IS_CELL_MASK_MATE_10_BIT_1((x)) || __IS_CELL_MASK_MATE_10_BIT_2((x)))

/*  ********************************************************************************************************** */

#define CELL_MASK_MATE_11_BIT_1                  0X0000000000400000
#define CELL_MASK_MATE_11_BIT_2                  0X0000000000800000
#define CELL_MASK_MATE_11_BIT_1_NEG              0XFFFFFFFFFFBFFFFF
#define CELL_MASK_MATE_11_BIT_2_NEG              0XFFFFFFFFFF7FFFFF

#define CELL_MASK_MATE_11_RESET                  0XFFFFFFFFFF3FFFFF
#define CELL_MASK_MATE_11_RESET_NEG              0X0000000000C00000

#define __UNSET_CELL_MASK_MATE_11(x)             ((x) &= CELL_MASK_MATE_11_RESET)
#define __SET_CELL_MASK_MATE_11_SET_1(x)         ((x) |= CELL_MASK_MATE_11_BIT_1)    ; ((x) &= CELL_MASK_MATE_11_BIT_2_NEG)
#define __SET_CELL_MASK_MATE_11_SET_2(x)         ((x) &= CELL_MASK_MATE_11_BIT_1_NEG); ((x) |= CELL_MASK_MATE_11_BIT_2)
#define __SET_CELL_MASK_MATE_11_SET_3(x)         ((x) |= CELL_MASK_MATE_11_BIT_1)    ; ((x) |= CELL_MASK_MATE_11_BIT_2)

#define __IS_CELL_MASK_MATE_11_BIT_1(x)          ((x) & CELL_MASK_MATE_11_BIT_1)
#define __IS_CELL_MASK_MATE_11_BIT_2(x)          ((x) & CELL_MASK_MATE_11_BIT_2)
#define __IS_CELL_MASK_MATE_11(x)                (__IS_CELL_MASK_MATE_11_BIT_1((x)) || __IS_CELL_MASK_MATE_11_BIT_2((x)))

/*  ********************************************************************************************************** */

__USE__DEVICE__ __USE__INLINE__ bool CellMaskIsMate(unsigned long * out, int idx){
  switch(idx){
  case(0): return __IS_CELL_MASK_MATE_0(*out);
  case(1): return __IS_CELL_MASK_MATE_1(*out);
  case(2): return __IS_CELL_MASK_MATE_2(*out);
  case(3): return __IS_CELL_MASK_MATE_3(*out);
  case(4): return __IS_CELL_MASK_MATE_4(*out);
  case(5): return __IS_CELL_MASK_MATE_5(*out);
  case(6): return __IS_CELL_MASK_MATE_6(*out);
  case(7): return __IS_CELL_MASK_MATE_7(*out);
  case(8): return __IS_CELL_MASK_MATE_8(*out);
  case(9): return __IS_CELL_MASK_MATE_9(*out);
  case(10):return __IS_CELL_MASK_MATE_10(*out);
  case(11):return __IS_CELL_MASK_MATE_11(*out);

  case(12):return __IS_CELL_MASK_MATE_0_BIT_2(*out);
  case(13):return __IS_CELL_MASK_MATE_1_BIT_2(*out);
  case(14):return __IS_CELL_MASK_MATE_2_BIT_2(*out);
  case(15):return __IS_CELL_MASK_MATE_3_BIT_2(*out);
  case(16):return __IS_CELL_MASK_MATE_4_BIT_2(*out);
  case(17):return __IS_CELL_MASK_MATE_5_BIT_2(*out);
  case(18):return __IS_CELL_MASK_MATE_6_BIT_2(*out);
  case(19):return __IS_CELL_MASK_MATE_7_BIT_2(*out);
  case(20):return __IS_CELL_MASK_MATE_8_BIT_2(*out);
  case(21):return __IS_CELL_MASK_MATE_9_BIT_2(*out);
  case(22):return __IS_CELL_MASK_MATE_10_BIT_2(*out);
  case(23):return __IS_CELL_MASK_MATE_11_BIT_2(*out);

  case(24):return (__IS_CELL_MASK_MATE_0_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(25):return (__IS_CELL_MASK_MATE_1_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(26):return (__IS_CELL_MASK_MATE_2_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(27):return (__IS_CELL_MASK_MATE_3_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(28):return (__IS_CELL_MASK_MATE_4_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(29):return (__IS_CELL_MASK_MATE_5_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(30):return (__IS_CELL_MASK_MATE_6_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(31):return (__IS_CELL_MASK_MATE_7_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(32):return (__IS_CELL_MASK_MATE_8_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(33):return (__IS_CELL_MASK_MATE_9_BIT_1(*out))  && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(34):return (__IS_CELL_MASK_MATE_10_BIT_1(*out)) && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  case(35):return (__IS_CELL_MASK_MATE_11_BIT_1(*out)) && (__IS_CELL_MASK_MATE_0_BIT_2(*out));
  default:return false;
  }
}

__USE__DEVICE__ __USE__INLINE__ void CellMaskSetMate(unsigned long * out, int idx){
  switch(idx){
  case(0):__SET_CELL_MASK_MATE_0_SET_1(*out); break;
  case(1):__SET_CELL_MASK_MATE_1_SET_1(*out); break;
  case(2):__SET_CELL_MASK_MATE_2_SET_1(*out); break;
  case(3):__SET_CELL_MASK_MATE_3_SET_1(*out); break;
  case(4):__SET_CELL_MASK_MATE_4_SET_1(*out); break;
  case(5):__SET_CELL_MASK_MATE_5_SET_1(*out); break;
  case(6):__SET_CELL_MASK_MATE_6_SET_1(*out); break;
  case(7):__SET_CELL_MASK_MATE_7_SET_1(*out); break;
  case(8):__SET_CELL_MASK_MATE_8_SET_1(*out); break;
  case(9):__SET_CELL_MASK_MATE_9_SET_1(*out); break;
  case(10):__SET_CELL_MASK_MATE_10_SET_1(*out); break;
  case(11):__SET_CELL_MASK_MATE_11_SET_1(*out); break;

  case(12):__SET_CELL_MASK_MATE_0_SET_2(*out); break;
  case(13):__SET_CELL_MASK_MATE_1_SET_2(*out); break;
  case(14):__SET_CELL_MASK_MATE_2_SET_2(*out); break;
  case(15):__SET_CELL_MASK_MATE_3_SET_2(*out); break;
  case(16):__SET_CELL_MASK_MATE_4_SET_2(*out); break;
  case(17):__SET_CELL_MASK_MATE_5_SET_2(*out); break;
  case(18):__SET_CELL_MASK_MATE_6_SET_2(*out); break;
  case(19):__SET_CELL_MASK_MATE_7_SET_2(*out); break;
  case(20):__SET_CELL_MASK_MATE_8_SET_2(*out); break;
  case(21):__SET_CELL_MASK_MATE_9_SET_2(*out); break;
  case(22):__SET_CELL_MASK_MATE_10_SET_2(*out); break;
  case(23):__SET_CELL_MASK_MATE_11_SET_2(*out); break;

  case(24):__SET_CELL_MASK_MATE_0_SET_3(*out); break;
  case(25):__SET_CELL_MASK_MATE_1_SET_3(*out); break;
  case(26):__SET_CELL_MASK_MATE_2_SET_3(*out); break;
  case(27):__SET_CELL_MASK_MATE_3_SET_3(*out); break;
  case(28):__SET_CELL_MASK_MATE_4_SET_3(*out); break;
  case(29):__SET_CELL_MASK_MATE_5_SET_3(*out); break;
  case(30):__SET_CELL_MASK_MATE_6_SET_3(*out); break;
  case(31):__SET_CELL_MASK_MATE_7_SET_3(*out); break;
  case(32):__SET_CELL_MASK_MATE_8_SET_3(*out); break;
  case(33):__SET_CELL_MASK_MATE_9_SET_3(*out); break;
  case(34):__SET_CELL_MASK_MATE_10_SET_3(*out); break;
  case(35):__SET_CELL_MASK_MATE_11_SET_3(*out); break;
    default: break;
  }
}

__USE__DEVICE__ __USE__INLINE__ void CellMaskUnsetMate(unsigned long * out, int idx){
  switch(idx){
  case(0):__UNSET_CELL_MASK_MATE_0(*out); break;
  case(1):__UNSET_CELL_MASK_MATE_1(*out); break;
  case(2):__UNSET_CELL_MASK_MATE_2(*out); break;
  case(3):__UNSET_CELL_MASK_MATE_3(*out); break;
  case(4):__UNSET_CELL_MASK_MATE_4(*out); break;
  case(5):__UNSET_CELL_MASK_MATE_5(*out); break;
  case(6):__UNSET_CELL_MASK_MATE_6(*out); break;
  case(7):__UNSET_CELL_MASK_MATE_7(*out); break;
  case(8):__UNSET_CELL_MASK_MATE_8(*out); break;
  case(9):__UNSET_CELL_MASK_MATE_9(*out); break;
  case(10):__UNSET_CELL_MASK_MATE_10(*out); break;
  case(11):__UNSET_CELL_MASK_MATE_11(*out); break;

  case(12):__SET_CELL_MASK_MATE_0_SET_1(*out); break;
  case(13):__SET_CELL_MASK_MATE_1_SET_1(*out); break;
  case(14):__SET_CELL_MASK_MATE_2_SET_1(*out); break;
  case(15):__SET_CELL_MASK_MATE_3_SET_1(*out); break;
  case(16):__SET_CELL_MASK_MATE_4_SET_1(*out); break;
  case(17):__SET_CELL_MASK_MATE_5_SET_1(*out); break;
  case(18):__SET_CELL_MASK_MATE_6_SET_1(*out); break;
  case(19):__SET_CELL_MASK_MATE_7_SET_1(*out); break;
  case(20):__SET_CELL_MASK_MATE_8_SET_1(*out); break;
  case(21):__SET_CELL_MASK_MATE_9_SET_1(*out); break;
  case(22):__SET_CELL_MASK_MATE_10_SET_1(*out); break;
  case(23):__SET_CELL_MASK_MATE_11_SET_1(*out); break;

  case(24):__SET_CELL_MASK_MATE_0_SET_2(*out); break;
  case(25):__SET_CELL_MASK_MATE_1_SET_2(*out); break;
  case(26):__SET_CELL_MASK_MATE_2_SET_2(*out); break;
  case(27):__SET_CELL_MASK_MATE_3_SET_2(*out); break;
  case(28):__SET_CELL_MASK_MATE_4_SET_2(*out); break;
  case(29):__SET_CELL_MASK_MATE_5_SET_2(*out); break;
  case(30):__SET_CELL_MASK_MATE_6_SET_2(*out); break;
  case(31):__SET_CELL_MASK_MATE_7_SET_2(*out); break;
  case(32):__SET_CELL_MASK_MATE_8_SET_2(*out); break;
  case(33):__SET_CELL_MASK_MATE_9_SET_2(*out); break;
  case(34):__SET_CELL_MASK_MATE_10_SET_2(*out); break;
  case(35):__SET_CELL_MASK_MATE_11_SET_2(*out); break;
  default: break;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates0(unsigned long mask){
  switch(CELL_MASK_MATE_0_RESET_NEG & mask){
    case(CELL_MASK_MATE_0_BIT_1): return 1;
    case(CELL_MASK_MATE_0_BIT_2): return 2;
    case(CELL_MASK_MATE_0_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates1(unsigned long mask){
  switch(CELL_MASK_MATE_1_RESET_NEG & mask){
    case(CELL_MASK_MATE_1_BIT_1): return 1;
    case(CELL_MASK_MATE_1_BIT_2): return 2;
    case(CELL_MASK_MATE_1_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates2(unsigned long mask){
  switch(CELL_MASK_MATE_2_RESET_NEG & mask){
    case(CELL_MASK_MATE_2_BIT_1): return 1;
    case(CELL_MASK_MATE_2_BIT_2): return 2;
    case(CELL_MASK_MATE_2_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates3(unsigned long mask){
  switch(CELL_MASK_MATE_3_RESET_NEG & mask){
    case(CELL_MASK_MATE_3_BIT_1): return 1;
    case(CELL_MASK_MATE_3_BIT_2): return 2;
    case(CELL_MASK_MATE_3_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates4(unsigned long mask){
  switch(CELL_MASK_MATE_4_RESET_NEG & mask){
    case(CELL_MASK_MATE_4_BIT_1): return 1;
    case(CELL_MASK_MATE_4_BIT_2): return 2;
    case(CELL_MASK_MATE_4_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates5(unsigned long mask){
  switch(CELL_MASK_MATE_5_RESET_NEG & mask){
    case(CELL_MASK_MATE_5_BIT_1): return 1;
    case(CELL_MASK_MATE_5_BIT_2): return 2;
    case(CELL_MASK_MATE_5_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates6(unsigned long mask){
  switch(CELL_MASK_MATE_6_RESET_NEG & mask){
    case(CELL_MASK_MATE_6_BIT_1): return 1;
    case(CELL_MASK_MATE_6_BIT_2): return 2;
    case(CELL_MASK_MATE_6_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates7(unsigned long mask){
  switch(CELL_MASK_MATE_7_RESET_NEG & mask){
    case(CELL_MASK_MATE_7_BIT_1): return 1;
    case(CELL_MASK_MATE_7_BIT_2): return 2;
    case(CELL_MASK_MATE_7_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates8(unsigned long mask){
  switch(CELL_MASK_MATE_8_RESET_NEG & mask){
    case(CELL_MASK_MATE_8_BIT_1): return 1;
    case(CELL_MASK_MATE_8_BIT_2): return 2;
    case(CELL_MASK_MATE_8_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates9(unsigned long mask){
  switch(CELL_MASK_MATE_9_RESET_NEG & mask){
    case(CELL_MASK_MATE_9_BIT_1): return 1;
    case(CELL_MASK_MATE_9_BIT_2): return 2;
    case(CELL_MASK_MATE_9_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates10(unsigned long mask){
  switch(CELL_MASK_MATE_10_RESET_NEG & mask){
    case(CELL_MASK_MATE_10_BIT_1): return 1;
    case(CELL_MASK_MATE_10_BIT_2): return 2;
    case(CELL_MASK_MATE_10_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates11(unsigned long mask){
  switch(CELL_MASK_MATE_11_RESET_NEG & mask){
    case(CELL_MASK_MATE_11_BIT_1): return 1;
    case(CELL_MASK_MATE_11_BIT_2): return 2;
    case(CELL_MASK_MATE_11_RESET_NEG): return 3;
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMates(unsigned long mask, int dir){
  switch(dir){
    case(0): return CellMaskGetCntMates0(mask);
    case(1): return CellMaskGetCntMates1(mask);
    case(2): return CellMaskGetCntMates2(mask);
    case(3): return CellMaskGetCntMates3(mask);
    case(4): return CellMaskGetCntMates4(mask);
    case(5): return CellMaskGetCntMates5(mask);
    case(6): return CellMaskGetCntMates6(mask);
    case(7): return CellMaskGetCntMates7(mask);
    case(8): return CellMaskGetCntMates8(mask);
    case(9): return CellMaskGetCntMates9(mask);
    case(10): return CellMaskGetCntMates10(mask);
    case(11): return CellMaskGetCntMates11(mask);
    default: return 0;
  }
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMatesAll(unsigned long mask){
  int result = 0;
  result += CellMaskGetCntMates0(mask);
  result += CellMaskGetCntMates1(mask);
  result += CellMaskGetCntMates2(mask);
  result += CellMaskGetCntMates3(mask);
  result += CellMaskGetCntMates4(mask);
  result += CellMaskGetCntMates5(mask);
  result += CellMaskGetCntMates6(mask);
  result += CellMaskGetCntMates7(mask);
  result += CellMaskGetCntMates8(mask);
  result += CellMaskGetCntMates9(mask);
  result += CellMaskGetCntMates10(mask);
  result += CellMaskGetCntMates11(mask);
  return result;
}

__USE__DEVICE__ __USE__INLINE__ int CellMaskGetCntMatesDir(unsigned long mask){
  int result = 0;
  if(__IS_CELL_MASK_MATE_0(mask))result++;
  if(__IS_CELL_MASK_MATE_1(mask))result++;
  if(__IS_CELL_MASK_MATE_2(mask))result++;
  if(__IS_CELL_MASK_MATE_3(mask))result++;
  if(__IS_CELL_MASK_MATE_4(mask))result++;
  if(__IS_CELL_MASK_MATE_5(mask))result++;
  if(__IS_CELL_MASK_MATE_6(mask))result++;
  if(__IS_CELL_MASK_MATE_7(mask))result++;
  if(__IS_CELL_MASK_MATE_8(mask))result++;
  if(__IS_CELL_MASK_MATE_9(mask))result++;
  if(__IS_CELL_MASK_MATE_10(mask))result++;
  if(__IS_CELL_MASK_MATE_11(mask))result++;
  return result;
}

__USE__DEVICE__ __USE__INLINE__ void CellMaskFillArrayCntMates(unsigned long mask, int* out, int cnt){
  int idx = cnt;
  if(idx <= 0)return; out[cnt - idx--] = CellMaskGetCntMates0(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates1(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates2(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates3(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates4(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates5(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates6(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates7(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates8(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates9(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates10(mask);
  if(idx == 0)return; out[cnt - idx--] = CellMaskGetCntMates11(mask);
}

/** ********************************************************************************************************** */
/*  ********************************************************************************************************** */
/** ********************************************************************************************************** */

__USE__DEVICE__ __USE__INLINE__ bool CellMaskHasAllMates(unsigned long mask){
  switch(CellMaskGetCrystalStruct(mask)){
    case(BCC): return __IS_CELL_MASK_MATE_0(mask) &&
                      __IS_CELL_MASK_MATE_1(mask) &&
                      __IS_CELL_MASK_MATE_2(mask) &&
                      __IS_CELL_MASK_MATE_3(mask) &&
                      __IS_CELL_MASK_MATE_4(mask) &&
                      __IS_CELL_MASK_MATE_5(mask) &&
                      __IS_CELL_MASK_MATE_6(mask) &&
                      __IS_CELL_MASK_MATE_7(mask);
    case(FCC): return __IS_CELL_MASK_MATE_0(mask)  &&
                      __IS_CELL_MASK_MATE_1(mask)  &&
                      __IS_CELL_MASK_MATE_2(mask)  &&
                      __IS_CELL_MASK_MATE_3(mask)  &&
                      __IS_CELL_MASK_MATE_4(mask)  &&
                      __IS_CELL_MASK_MATE_5(mask)  &&
                      __IS_CELL_MASK_MATE_6(mask)  &&
                      __IS_CELL_MASK_MATE_7(mask)  &&
                      __IS_CELL_MASK_MATE_8(mask)  &&
                      __IS_CELL_MASK_MATE_9(mask)  &&
                      __IS_CELL_MASK_MATE_10(mask) &&
                      __IS_CELL_MASK_MATE_11(mask) ;
  default: return false;
  }
}

/** ********************************************************************************************************** */
/*  ********************************************************************************************************** */
/** ********************************************************************************************************** */

#define CELL_MASK_SCENE_BORDER                  0X0000000100000000
#define CELL_MASK_PART_DB_BORDER                0X0000000200000000
#define CELL_MASK_BORDER_ANY                    0X0000000300000000

#define CELL_MASK_NEAR_BUG                      0X0000000400000000
#define CELL_MASK_FAR_BUG                       0X0000000800000000
#define CELL_MASK_BUG_VAC                       0X0000001000000000
#define CELL_MASK_BUG_INT                       0X0000002000000000
#define CELL_MASK_BUG_INT_BIG                   0X0000004000000000
#define CELL_MASK_BUG_INT_ANY                   0X0000006000000000
#define CELL_MASK_BUG_SUB                       0X0000008000000000
#define CELL_MASK_BUG_LINE_DIS                  0X0000010000000000
#define CELL_MASK_BUG_SC_DIS                    0X0000020000000000
#define CELL_MASK_BUG_UNDEF1                    0X0000040000000000
#define CELL_MASK_BUG_UNDEF2                    0X0000080000000000
#define CELL_MASK_BUG_UNDEF3                    0X0000100000000000
#define CELL_MASK_BUG_UNDEF4                    0X0000200000000000
#define CELL_MASK_BUG_UNDEF5                    0X0000400000000000
#define CELL_MASK_BUG_UNDEF6                    0X0000800000000000

#define CELL_MASK_BUG_ANY                       0X0000FFF000000000
#define CELL_MASK_BUG_OR_NEAR                   0X0000FFF400000000
#define CELL_MASK_BUG_OR_NEAR_OR_FAR            0X0000FFFC00000000

#define __SET_CELL_MASK_SCENE_BORDER(x)         ((x) |= CELL_MASK_SCENE_BORDER)
#define __SET_CELL_MASK_PART_DB_BORDER(x)       ((x) |= CELL_MASK_PART_DB_BORDER)
#define __SET_CELL_MASK_NEAR_BUG(x)             ((x) |= CELL_MASK_NEAR_BUG)
#define __SET_CELL_MASK_FAR_BUG(x)              ((x) |= CELL_MASK_FAR_BUG)
#define __SET_CELL_MASK_BUG_VAC(x)              ((x) |= CELL_MASK_BUG_VAC)
#define __SET_CELL_MASK_BUG_INT(x)              ((x) |= CELL_MASK_BUG_INT)
#define __SET_CELL_MASK_BUG_INT_BIG(x)          ((x) |= CELL_MASK_BUG_INT_BIG)
#define __SET_CELL_MASK_BUG_SUB(x)              ((x) |= CELL_MASK_BUG_SUB)
#define __SET_CELL_MASK_BUG_LINE_DIS(x)         ((x) |= CELL_MASK_BUG_LINE_DIS)
#define __SET_CELL_MASK_BUG_SC_DIS(x)           ((x) |= CELL_MASK_BUG_SC_DIS)


#define __UNSET_CELL_MASK_SCENE_BORDER(x)       ((x) &= (~CELL_MASK_SCENE_BORDER))
#define __UNSET_CELL_MASK_PART_DB_BORDER(x)     ((x) &= (~CELL_MASK_PART_DB_BORDER))
#define __UNSET_CELL_MASK_NEAR_BUG(x)           ((x) &= (~CELL_MASK_NEAR_BUG))
#define __UNSET_CELL_MASK_FAR_BUG(x)            ((x) &= (~CELL_MASK_FAR_BUG))
#define __UNSET_CELL_MASK_BUG_VAC(x)            ((x) &= (~CELL_MASK_BUG_VAC))
#define __UNSET_CELL_MASK_BUG_INT(x)            ((x) &= (~CELL_MASK_BUG_INT))
#define __UNSET_CELL_MASK_BUG_INT_BIG(x)        ((x) &= (~CELL_MASK_BUG_INT_BIG))
#define __UNSET_CELL_MASK_BUG_SUB(x)            ((x) &= (~CELL_MASK_BUG_SUB))
#define __UNSET_CELL_MASK_BUG_LINE_DIS(x)       ((x) &= (~CELL_MASK_BUG_LINE_DIS))
#define __UNSET_CELL_MASK_BUG_SC_DIS(x)         ((x) &= (~CELL_MASK_BUG_SC_DIS))


#define __IS_CELL_MASK_SCENE_BORDER(x)          ((x) & CELL_MASK_SCENE_BORDER)
#define __IS_CELL_MASK_PART_DB_BORDER(x)        ((x) & CELL_MASK_PART_DB_BORDER)
#define __IS_CELL_MASK_NEAR_BUG(x)              ((x) & CELL_MASK_NEAR_BUG)
#define __IS_CELL_MASK_FAR_BUG(x)               ((x) & CELL_MASK_FAR_BUG)
#define __IS_CELL_MASK_BUG_VAC(x)               ((x) & CELL_MASK_BUG_VAC)
#define __IS_CELL_MASK_BUG_INT(x)               ((x) & CELL_MASK_BUG_INT)
#define __IS_CELL_MASK_BUG_INT_BIG(x)           ((x) & CELL_MASK_BUG_INT_BIG)
#define __IS_CELL_MASK_BUG_INT_ANY(x)           ((x) & CELL_MASK_BUG_INT_ANY)
#define __IS_CELL_MASK_BUG_SUB(x)               ((x) & CELL_MASK_BUG_SUB)
#define __IS_CELL_MASK_BUG_LINE_DIS(x)          ((x) & CELL_MASK_BUG_LINE_DIS)
#define __IS_CELL_MASK_BUG_SC_DIS(x)            ((x) & CELL_MASK_BUG_SC_DIS)

#define __IS_CELL_MASK_BORDER_ANY(x)            ((x) & CELL_MASK_BORDER_ANY)
#define __IS_CELL_MASK_BUG_ANY(x)               ((x) & CELL_MASK_BUG_ANY)
#define __IS_CELL_MASK_BUG_OR_NEAR(x)           ((x) & CELL_MASK_BUG_OR_NEAR)
#define __IS_CELL_MASK_BUG_OR_NEAR_OR_FAR(x)    ((x) & CELL_MASK_BUG_OR_NEAR_OR_FAR)

__USE__DEVICE__ __USE__INLINE__ void CellMaskSetBug(unsigned long* mask, BugType bug){
  switch(bug){
    case(VACANCY):     __SET_CELL_MASK_BUG_VAC(*mask);       return;
//    case(INTERSTITIAL):__SET_CELL_MASK_BUG_INT(*mask);       return;
    case(SUBSTITUTION):__SET_CELL_MASK_BUG_SUB(*mask);       return;
  //case(DISLOCATION__BugType): __SET_CELL_MASK_BUG_LINE_DIS(*mask);  return;
  }
}

/** ********************************************************************************************************** */
/*  ********************************************************************************************************** */
/** ********************************************************************************************************** */

#define CNT_DIR_BCC     8
#define CNT_DIR_FCC     12

__USE__DEVICE__ __USE__INLINE__ int CellMateReverseDir(int dir){ return dir % 2 == 0 ? dir + 1 : dir - 1; }

__USE__DEVICE__ __USE__INLINE__ void SetLaticePositionsBCC(int dir, float3 * out, float3 * pos,
                                                           float aX, float aY, float aZ,
                                                           float bX, float bY, float bZ,
                                                           float gX, float gY, float gZ){
  switch(dir){
  case(PA_PB_PG):
    out->x = pos->x + aX*0.5f + bX*0.5f + gX*0.5f;
    out->y = pos->y + aY*0.5f + bY*0.5f + gY*0.5f;
    out->z = pos->z + aZ*0.5f + bZ*0.5f + gZ*0.5f;
    break;
  case(MA_MB_MG):
    out->x = pos->x - aX*0.5f - bX*0.5f - gX*0.5f;
    out->y = pos->y - aY*0.5f - bY*0.5f - gY*0.5f;
    out->z = pos->z - aZ*0.5f - bZ*0.5f - gZ*0.5f;
    break;
  case(PA_PB_MG):
    out->x = pos->x + aX*0.5f + bX*0.5f - gX*0.5f;
    out->y = pos->y + aY*0.5f + bY*0.5f - gY*0.5f;
    out->z = pos->z + aZ*0.5f + bZ*0.5f - gZ*0.5f;
    break;
  case(MA_MB_PG):
    out->x = pos->x - aX*0.5f - bX*0.5f + gX*0.5f;
    out->y = pos->y - aY*0.5f - bY*0.5f + gY*0.5f;
    out->z = pos->z - aZ*0.5f - bZ*0.5f + gZ*0.5f;
    break;
  case(PA_MB_PG):
    out->x = pos->x + aX*0.5f - bX*0.5f + gX*0.5f;
    out->y = pos->y + aY*0.5f - bY*0.5f + gY*0.5f;
    out->z = pos->z + aZ*0.5f - bZ*0.5f + gZ*0.5f;
    break;
  case(MA_PB_MG):
    out->x = pos->x - aX*0.5f + bX*0.5f - gX*0.5f;
    out->y = pos->y - aY*0.5f + bY*0.5f - gY*0.5f;
    out->z = pos->z - aZ*0.5f + bZ*0.5f - gZ*0.5f;
    break;
  case(PA_MB_MG):
    out->x = pos->x + aX*0.5f - bX*0.5f - gX*0.5f;
    out->y = pos->y + aY*0.5f - bY*0.5f - gY*0.5f;
    out->z = pos->z + aZ*0.5f - bZ*0.5f - gZ*0.5f;
    break;
  case(MA_PB_PG):
    out->x = pos->x - aX*0.5f + bX*0.5f + gX*0.5f;
    out->y = pos->y - aY*0.5f + bY*0.5f + gY*0.5f;
    out->z = pos->z - aZ*0.5f + bZ*0.5f + gZ*0.5f;
    break;
  default:
    out->x = pos->x;
    out->y = pos->y;
    out->z = pos->z;
    break;
  }
}


__USE__DEVICE__ __USE__INLINE__ void SetLaticePositionsFCC(int dir, float3 * out, float3 * pos,
                                                           float aX, float aY, float aZ,
                                                           float bX, float bY, float bZ,
                                                           float gX, float gY, float gZ){
  switch(dir){
  case(PA_PB):
    out->x = pos->x + aX*0.5f + bX*0.5f;
    out->y = pos->y + aY*0.5f + bY*0.5f;
    out->z = pos->z + aZ*0.5f + bZ*0.5f;
    break;
  case(MA_MB):
    out->x = pos->x - aX*0.5f - bX*0.5f;
    out->y = pos->y - aY*0.5f - bY*0.5f;
    out->z = pos->z - aZ*0.5f - bZ*0.5f;
    break;
  case(PA_MB):
    out->x = pos->x + aX*0.5f - bX*0.5f;
    out->y = pos->y + aY*0.5f - bY*0.5f;
    out->z = pos->z + aZ*0.5f - bZ*0.5f;
    break;
  case(MA_PB):
    out->x = pos->x - aX*0.5f + bX*0.5f;
    out->y = pos->y - aY*0.5f + bY*0.5f;
    out->z = pos->z - aZ*0.5f + bZ*0.5f;
    break;
  case(PA_PG):
    out->x = pos->x + aX*0.5f + gX*0.5f;
    out->y = pos->y + aY*0.5f + gY*0.5f;
    out->z = pos->z + aZ*0.5f + gZ*0.5f;
    break;
  case(MA_MG):
    out->x = pos->x - aX*0.5f - gX*0.5f;
    out->y = pos->y - aY*0.5f - gY*0.5f;
    out->z = pos->z - aZ*0.5f - gZ*0.5f;
    break;
  case(PA_MG):
    out->x = pos->x + aX*0.5f - gX*0.5f;
    out->y = pos->y + aY*0.5f - gY*0.5f;
    out->z = pos->z + aZ*0.5f - gZ*0.5f;
    break;
  case(MA_PG):
    out->x = pos->x - aX*0.5f + gX*0.5f;
    out->y = pos->y - aY*0.5f + gY*0.5f;
    out->z = pos->z - aZ*0.5f + gZ*0.5f;
    break;
  case(PB_PG):
    out->x = pos->x + bX*0.5f + gX*0.5f;
    out->y = pos->y + bY*0.5f + gY*0.5f;
    out->z = pos->z + bZ*0.5f + gZ*0.5f;
    break;
  case(MB_MG):
    out->x = pos->x - bX*0.5f - gX*0.5f;
    out->y = pos->y - bY*0.5f - gY*0.5f;
    out->z = pos->z - bZ*0.5f - gZ*0.5f;
    break;
  case(PB_MG):
    out->x = pos->x + bX*0.5f - gX*0.5f;
    out->y = pos->y + bY*0.5f - gY*0.5f;
    out->z = pos->z + bZ*0.5f - gZ*0.5f;
    break;
  case(MB_PG):
    out->x = pos->x - bX*0.5f + gX*0.5f;
    out->y = pos->y - bY*0.5f + gY*0.5f;
    out->z = pos->z - bZ*0.5f + gZ*0.5f;
    break;
  default:
    out->x = pos->x;
    out->y = pos->y;
    out->z = pos->z;
    break;
  }
}

#endif // ___CUDA_CELL_MASK_H___
