#ifndef PROJECT_H
#define PROJECT_H

#include "cellmask.h"

enum BorderVelocityType
{
  PRESS_X, PULL_X,
  PRESS_Y, PULL_Y,
  PRESS_Z, PULL_Z
};

struct VecF
{
  float x, y, z;

  VecF(float v = 0.0f) : x(v), y(v), z(v){}
  VecF(float _x, float _y, float _z) : x(_x), y(_y), z(_z){}

  inline float length() { return sqrt(x*x + y*y + z*z); }

  inline void set(float _x, float _y, float _z) { x = _x; y = _y; z = _z;}
  inline void add(float _x, float _y, float _z) { x = x + _x; y = y + _y; z = z + _z;}
  inline void mul(float val){  x = x*val; y = y*val; z = z*val; }
  inline void div(float val){  x = x/val; y = y/val; z = z/val; }
  inline void normalize(){ div(length()); }
  inline void setLength(float length){ normalize(); mul(length); }
  inline void setNormal(VecF & v1, VecF & v2) {
    x = v1.y * v2.z - v1.z * v2.y;
    y = v1.z * v2.x - v1.x * v2.z;
    z = v1.x * v2.y - v1.y * v2.x;
  }
  inline QString toString(){ return QString("[%1,%2,%3]").arg(x).arg(y).arg(z); }
};

struct ProjectDataF
{
  int             count, countVac, seedIdx, countBorder;
  VecF            * pos;
  VecF            * stres;
  CellAttribute   * attr;
  CellMate        * mate;
  int             * wallIds;

  ProjectDataF(int _count, int _seedIdx);
  ~ProjectDataF();
  bool storeMol2File(QString fileNameBase, QString name, int stepId, bool storeBonds = false);
};

struct ProjectF
{
  float               wallX0, wallX1;
  float               wallY0, wallY1;
  float               wallZ0, wallZ1;

  BorderVelocityType  wallMoveType;
  int                 wallMoveSteps;
  float               wallMoveLength;

  CrystalMatrixType   matrixType;

  float crystalA, crystalLJE, crystalLJD;
  VecF  seed, alpha, beta, gama;

  bool  useVacancy;
  float vacancyRatio;

  bool  useSubstitution;
  float substitutionRatio;
  float substitutionA, substitutionLJE, substitutionLJD;

  QString name;

  ProjectF();
};

#endif // PROJECT_H
