#include "stabil.h"
Stabilisation::Stabilisation(){}

bool Stabilisation::process(ProjectF * prj, ProjectDataF * data, float stopValue, int totalLoops){

  for(int step = 0 ; step < totalLoops ; step++)
  {
    float tmp, max = 0.0f;

    for(int idx = 0 ; idx < data->count ; idx ++){
      ulong mask = data->attr[idx].mask;
      if(mask == CELL_MASK_NULL)continue;
      if(__IS_CELL_MASK_SCENE_BORDER(mask)){  tmp = stabilBorder(idx, prj, data); if(max < tmp)max = tmp; }
      else if(__IS_CELL_MASK_BUG_VAC(mask)){  tmp = stabilVac(idx, prj, data);    if(max < tmp)max = tmp; }
      else if(__IS_CELL_MASK_BUG_SUB(mask)){  tmp = stabilSub(idx, prj, data);    if(max < tmp)max = tmp; }
      else if(__IS_CELL_MASK_NEAR_BUG(mask)){ tmp = stabilNear(idx, prj, data);   if(max < tmp)max = tmp; }
      else{                                   tmp = stabilBase(idx, prj, data);   if(max < tmp)max = tmp; }
    }

    //qDebug() << max;

    for(int idx = 0 ; idx < data->count ; idx ++){
      data->pos[idx].x = data->stres[idx].x;
      data->pos[idx].y = data->stres[idx].y;
      data->pos[idx].z = data->stres[idx].z;
    }
    if(max < stopValue)return true;
  }
  return false;
}

float Stabilisation::stabilBorder(int idx, ProjectF * /*prj*/, ProjectDataF * data){
  data->stres[idx].x = data->pos[idx].x;
  data->stres[idx].y = data->pos[idx].y;
  data->stres[idx].z = data->pos[idx].z;
  return 0.0f;
}

float Stabilisation::stabilNear(int idx, ProjectF * prj, ProjectDataF * data){

  float foundMatesCount = 0.0f;
  VecF inPos;
  VecF outPos;

  inPos.set(data->pos[idx].x, data->pos[idx].y, data->pos[idx].z);
  outPos.set(0.0f, 0.0f, 0.0f);

  for (int mateIdx = 0 ; mateIdx < MAX_MATES ; mateIdx++)
  {
    VecF matePos, mateOptPos;
    int mateId = data->mate[idx].m1[mateIdx];
    if(mateId == 0)continue;

    ulong mateMask = data->attr[mateId].mask;
    if(mateMask == CELL_MASK_NULL)continue;

    matePos.set(data->pos[mateId].x, data->pos[mateId].y, data->pos[mateId].z);

    if(__IS_CELL_MASK_BUG_VAC(mateMask))
    {
      outPos.x += matePos.x - inPos.x;
      outPos.y += matePos.y - inPos.y;
      outPos.z += matePos.z - inPos.z;
    }
    else if(__IS_CELL_MASK_BUG_SUB(mateMask))
    {
      VecF a, b, g;
      a.set(prj->alpha.x, prj->alpha.y, prj->alpha.z);
      b.set(prj->beta.x, prj->beta.y, prj->beta.z);
      g.set(prj->gama.x, prj->gama.y, prj->gama.z);
      a.setLength(prj->substitutionA);
      b.setLength(prj->substitutionA);
      g.setLength(prj->substitutionA);
      prepareMateOptPos(&mateOptPos, prj->matrixType, a, b, g, inPos, mateIdx);
      outPos.x += matePos.x - mateOptPos.x;
      outPos.y += matePos.y - mateOptPos.y;
      outPos.z += matePos.z - mateOptPos.z;
    }else{
      prepareMateOptPos(&mateOptPos, prj->matrixType, prj->alpha, prj->beta, prj->gama, inPos, mateIdx);
      outPos.x += matePos.x - mateOptPos.x;
      outPos.y += matePos.y - mateOptPos.y;
      outPos.z += matePos.z - mateOptPos.z;
    }

    foundMatesCount += 1.0f;
  }

  outPos.div(foundMatesCount);

  data->stres[idx].x = inPos.x + outPos.x;
  data->stres[idx].y = inPos.y + outPos.y;
  data->stres[idx].z = inPos.z + outPos.z;

  return outPos.length();

}

float Stabilisation::stabilVac(int idx, ProjectF * /*prj*/, ProjectDataF * data){

  float foundMatesCount = 0.0f;
  VecF outPos;

  outPos.set(0.0f, 0.0f, 0.0f);

  for (int mateIdx = 0 ; mateIdx < MAX_MATES ; mateIdx++)
  {
    int mateId = data->mate[idx].m1[mateIdx];
    if(mateId == 0)continue;
    ulong mateMask = data->attr[mateId].mask;
    if(mateMask == CELL_MASK_NULL)continue;
    outPos.add(data->pos[mateId].x, data->pos[mateId].y, data->pos[mateId].z);
    foundMatesCount += 1.0f;
  }
  outPos.div(foundMatesCount);
  data->stres[idx].x = outPos.x;
  data->stres[idx].y = outPos.y;
  data->stres[idx].z = outPos.z;
  return 0.0f;
}

float Stabilisation::stabilBase(int idx, ProjectF * prj, ProjectDataF * data){

  float foundMatesCount = 0.0f;
  VecF inPos;
  VecF outPos;

  inPos.set(data->pos[idx].x, data->pos[idx].y, data->pos[idx].z);
  outPos.set(0.0f, 0.0f, 0.0f);

  for (int mateIdx = 0 ; mateIdx < MAX_MATES ; mateIdx++)
  {
    VecF matePos, mateOptPos;
    int mateId = data->mate[idx].m1[mateIdx];
    if(mateId == 0)continue;

    ulong mateMask = data->attr[mateId].mask;
    if(mateMask == CELL_MASK_NULL)continue;
    matePos.set(data->pos[mateId].x, data->pos[mateId].y, data->pos[mateId].z);

    prepareMateOptPos(&mateOptPos, prj->matrixType, prj->alpha, prj->beta, prj->gama, inPos, mateIdx);

    outPos.x += matePos.x - mateOptPos.x;
    outPos.y += matePos.y - mateOptPos.y;
    outPos.z += matePos.z - mateOptPos.z;

    foundMatesCount += 1.0f;
  }

  outPos.div(foundMatesCount);

  data->stres[idx].x = inPos.x + outPos.x;
  data->stres[idx].y = inPos.y + outPos.y;
  data->stres[idx].z = inPos.z + outPos.z;

  return outPos.length();
}

float Stabilisation::stabilSub(int idx, ProjectF * prj, ProjectDataF * data){

  float foundMatesCount = 0.0f;
  VecF inPos;
  VecF outPos;
  VecF a, b, g;

  a.set(prj->alpha.x, prj->alpha.y, prj->alpha.z);
  b.set(prj->beta.x, prj->beta.y, prj->beta.z);
  g.set(prj->gama.x, prj->gama.y, prj->gama.z);

  a.setLength(prj->substitutionA);
  b.setLength(prj->substitutionA);
  g.setLength(prj->substitutionA);

  inPos.set(data->pos[idx].x, data->pos[idx].y, data->pos[idx].z);
  outPos.set(0.0f, 0.0f, 0.0f);

  for (int mateIdx = 0 ; mateIdx < MAX_MATES ; mateIdx++)
  {
    VecF matePos, mateOptPos;
    int mateId = data->mate[idx].m1[mateIdx];
    if(mateId == 0)continue;
    ulong mateMask = data->attr[mateId].mask;
    if(mateMask == CELL_MASK_NULL)continue;
    matePos.set(data->pos[mateId].x, data->pos[mateId].y, data->pos[mateId].z);

    prepareMateOptPos(&mateOptPos, prj->matrixType, a, b, g, inPos, mateIdx);

    outPos.x += matePos.x - mateOptPos.x;
    outPos.y += matePos.y - mateOptPos.y;
    outPos.z += matePos.z - mateOptPos.z;

    foundMatesCount += 1.0f;
  }

  outPos.div(foundMatesCount);

  data->stres[idx].x = inPos.x + outPos.x;
  data->stres[idx].y = inPos.y + outPos.y;
  data->stres[idx].z = inPos.z + outPos.z;

  return outPos.length();
}

void Stabilisation::prepareMateOptPos(VecF * mateOptPos, CrystalMatrixType matrixType, VecF & alpha, VecF & beta, VecF & gama, VecF & inPos, int mateIdx)
{
  if(matrixType == BCC){
    if(mateIdx == PA_PB_PG) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f + beta.x / 2.0f + gama.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f + beta.y / 2.0f + gama.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f + beta.z / 2.0f + gama.z / 2.0f;
    } else if(mateIdx == MA_MB_MG) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f - beta.x / 2.0f - gama.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f - beta.y / 2.0f - gama.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f - beta.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == PA_PB_MG) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f + beta.x / 2.0f - gama.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f + beta.y / 2.0f - gama.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f + beta.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == MA_MB_PG) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f - beta.x / 2.0f + gama.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f - beta.y / 2.0f + gama.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f - beta.z / 2.0f + gama.z / 2.0f;
    } else if(mateIdx == PA_MB_PG) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f - beta.x / 2.0f + gama.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f - beta.y / 2.0f + gama.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f - beta.z / 2.0f + gama.z / 2.0f;
    } else if(mateIdx == MA_PB_MG) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f + beta.x / 2.0f - gama.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f + beta.y / 2.0f - gama.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f + beta.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == PA_MB_MG) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f - beta.x / 2.0f - gama.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f - beta.y / 2.0f - gama.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f - beta.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == MA_PB_PG) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f + beta.x / 2.0f + gama.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f + beta.y / 2.0f + gama.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f + beta.z / 2.0f + gama.z / 2.0f;
    }
  }else if(matrixType == FCC){
    if(mateIdx == PA_PB) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f + beta.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f + beta.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f + beta.z / 2.0f;
    } else if(mateIdx == MA_MB) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f - beta.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f - beta.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f - beta.z / 2.0f;
    } else if(mateIdx == PA_MB) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f - beta.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f - beta.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f - beta.z / 2.0f;
    } else if(mateIdx == MA_PB) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f + beta.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f + beta.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f + beta.z / 2.0f;
    } else

    if(mateIdx == PA_PG) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f + gama.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f + gama.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f + gama.z / 2.0f;
    } else if(mateIdx == MA_MG) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f - gama.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f - gama.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == PA_MG) {
      mateOptPos->x = inPos.x + alpha.x / 2.0f - gama.x / 2.0f;
      mateOptPos->y = inPos.y + alpha.y / 2.0f - gama.y / 2.0f;
      mateOptPos->z = inPos.z + alpha.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == MA_PG) {
      mateOptPos->x = inPos.x - alpha.x / 2.0f + gama.x / 2.0f;
      mateOptPos->y = inPos.y - alpha.y / 2.0f + gama.y / 2.0f;
      mateOptPos->z = inPos.z - alpha.z / 2.0f + gama.z / 2.0f;
    } else

    if(mateIdx == PB_PG) {
     mateOptPos->x = inPos.x + beta.x / 2.0f + gama.x / 2.0f;
     mateOptPos->y = inPos.y + beta.y / 2.0f + gama.y / 2.0f;
     mateOptPos->z = inPos.z + beta.z / 2.0f + gama.z / 2.0f;
    } else if(mateIdx == MB_MG) {
     mateOptPos->x = inPos.x - beta.x / 2.0f - gama.x / 2.0f;
     mateOptPos->y = inPos.y - beta.y / 2.0f - gama.y / 2.0f;
     mateOptPos->z = inPos.z - beta.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == PB_MG) {
     mateOptPos->x = inPos.x + beta.x / 2.0f - gama.x / 2.0f;
     mateOptPos->y = inPos.y + beta.y / 2.0f - gama.y / 2.0f;
     mateOptPos->z = inPos.z + beta.z / 2.0f - gama.z / 2.0f;
    } else if(mateIdx == MB_PG) {
     mateOptPos->x = inPos.x - beta.x / 2.0f + gama.x / 2.0f;
     mateOptPos->y = inPos.y - beta.y / 2.0f + gama.y / 2.0f;
     mateOptPos->z = inPos.z - beta.z / 2.0f + gama.z / 2.0f;
    }
  }
}

